from schemdraw import Drawing
from schemdraw import elements as elm

from .input import Input


def draw_void():
    drawing = Drawing()

    transistor = drawing.add(elm.transistors.BjtNpn().right())
    drawing += elm.Line().up()
    drawing += elm.Line().left().length(drawing.unit/2)

    drawing.push()
    drawing += elm.Dot()
    drawing += elm.SourceV().up().label('Vcc').reverse()
    drawing += elm.Ground().left()

    drawing.pop()
    drawing += elm.Line().left().length(drawing.unit/2)
    drawing += elm.Resistor().down().label('Rb')
    drawing += elm.Line().down().length(0.24*drawing.unit)
    drawing += elm.Dot()

    drawing.push()
    drawing += elm.Line().to(transistor.base)

    drawing.pop()
    drawing += elm.Line().left().length(drawing.unit/4)
    drawing += elm.Capacitor2().label('C1').left()
    drawing += elm.Line().left().length(drawing.unit/4)
    drawing += elm.Dot().label('Vi')

    drawing += elm.Resistor().down().label('Re').at(transistor.emitter)
    drawing += elm.Ground()

    drawing += elm.Dot().at(transistor.emitter)
    drawing += elm.Capacitor2().label('C2').right()
    drawing += elm.Dot().label('Vo')

    return drawing


def draw(amplifier_input: Input):
    Vcc = str(round(amplifier_input.Vcc, 2))
    Rb = str(round(amplifier_input.Rb, 2))
    Re = str(round(amplifier_input.Re, 2))

    drawing = Drawing()

    transistor = drawing.add(elm.transistors.BjtNpn().right())
    drawing += elm.Line().up()
    drawing += elm.Line().left().length(drawing.unit/2)

    drawing.push()
    drawing += elm.Dot()
    drawing += elm.SourceV().up().label(Vcc).reverse()
    drawing += elm.Ground().left()

    drawing.pop()
    drawing += elm.Line().left().length(drawing.unit/2)
    drawing += elm.Resistor().down().label(Rb)
    drawing += elm.Line().down().length(0.24*drawing.unit)
    drawing += elm.Dot()

    drawing.push()
    drawing += elm.Line().to(transistor.base)

    drawing.pop()
    drawing += elm.Line().left().length(drawing.unit/4)
    drawing += elm.Capacitor2().label('C1').left()
    drawing += elm.Line().left().length(drawing.unit/4)
    drawing += elm.Dot().label('Vi')

    drawing += elm.Resistor().down().label(Re).at(transistor.emitter)
    drawing += elm.Ground()

    drawing += elm.Dot().at(transistor.emitter)
    drawing += elm.Capacitor2().label('C2').right()
    drawing += elm.Dot().label('Vo')

    return drawing



class Result:
    V_load: complex
    I: complex

    def __init__(self, V_load: complex, I: complex):
        self.V_load = V_load
        self.I = I

    def __getitem__(self, item: int) -> complex:
        if item == 0:
            return self.V_load
        elif item == 1:
            return self.I
        else:
            raise IndexError("Index out of range")


def check_max_error(V_load: complex, V_load_old: complex, max_error: float) -> bool:
    real_error = abs(abs(V_load) - abs(V_load_old))
    imag_error = abs(V_load.imag - V_load_old.imag)
    if real_error <= max_error or imag_error <= max_error:
        return True
    return False


def constant_current(S_load: complex, Z_line: complex, V: complex) -> Result:
    V_load = V
    I = (S_load/V_load).conjugate()
    V_line = Z_line*I
    V_load = V - V_line

    return Result(V_load, I)


def constant_power(S_load: complex, Z_line: complex, V: complex, max_error: float) -> Result:
    V_load = V
    V_load_old: complex = max_error + 1

    while True:
        I = (S_load.conjugate()/V_load)
        V_line = Z_line*I
        V_load = V - V_line

        if check_max_error(V_load, V_load_old, max_error):
            break
        V_load_old = V_load

    I = S_load.conjugate()/V_load
    return Result(V_load, I)


def constant_impedance(S_load: complex, Z_line: complex, V: complex, max_error: float) -> Result:
    V_load = V
    V_load_old: complex = max_error + 1

    while True:
        Z_load = (V**2/S_load).conjugate()
        I = V_load/Z_load
        V_line = Z_line*I
        V_load = V - V_line

        if check_max_error(V_load, V_load_old, max_error):
            break
        V_load_old = V_load

    Z_load = (V**2/S_load).conjugate()
    I = V_load/Z_load
    return Result(V_load, I)



from typing import List

from pyelectric.electronic.amplifier.input import Input as _Input


class Input(_Input):
    Vdd: float
    R1: float
    R2: float
    Rd: float
    Rs: float
    Vp: float
    Idss: float
    rd: float

    def __init__(self, Vdd: float, R1: float, R2: float, Rd: float, Rs: float, Vp: float, Idss: float, rd: float) -> None:
        self.Vdd = Vdd
        self.R1 = R1
        self.R2 = R2
        self.Rd = Rd
        self.Rs = Rs
        self.Vp = Vp
        self.Idss = Idss
        self.rd = rd

    @staticmethod
    def get_parameter_names() -> List[str]:
        return ['Vdd', 'Vp', 'Idss', 'rd', 'R1', 'R2', 'Rd', 'Rs']



from . import Bar


class LoadBar(Bar):
    active_power: float
    reactive_power: float

    def __init__(self, name: str, active_power: float, reactive_power: float):
        super().__init__(name)
        self.active_power = active_power
        self.reactive_power = reactive_power
        self.power = self.active_power + 1j * self.reactive_power



from pyelectric.electronic.amplifier import Amplifier as _Amplifier

from . import drawing
from .input import Input
from .output import Output


class Amplifier(_Amplifier):
    input: Input = Input(0, 0, 0, 0, 0, 0, 0)
    output: Output = Output(0, 0, 0, 0)

    def __init__(self, *args, **kwargs) -> None:
        super().__init__()
        self.input = Input(*args, **kwargs)

    def __call__(self) -> Output:
        Vcc = self.input.Vcc
        beta = self.input.beta
        ro = self.input.ro
        R1 = self.input.R1
        R2 = self.input.R2
        Rc = self.input.Rc
        Re = self.input.Re

        R_line = 1/((1/R1) + (1/R2))
        Vb = Vcc*R2/(R1 + R2)
        Vbe = 0.7
        Ve = Vb - Vbe
        Ieq = Ve/Re

        Zo = 1/((1/ro) + (1/Rc))
        re = 26e-3/Ieq
        Zi = 1/((1/R_line) + (1/(beta*re)))
        Avnl = -Zo/re

        return Output(re, Zi, Zo, Avnl)

    @staticmethod
    def draw_void():
        return drawing.draw_void()

    def draw(self):
        return drawing.draw(self.input)



from .bar import Bar


class Line:
    admittance: float
    bar1: Bar
    bar2: Bar

    def __init__(self, bar1: Bar, bar2: Bar, *, admittance: complex = None, impedance: complex = None):
        assert admittance is not None or impedance is not None
        self.bar1 = bar1
        self.bar2 = bar2
        if admittance is not None:
            self.admittance = admittance
        else:
            self.admittance = 1/impedance



def complex_to_str(c: complex) -> str:
    return f'{c.real:.4e} + {c.imag:.4e}j'


class Bar:
    name: str
    voltage: complex = 0 + 0j
    power: complex = 0 + 0j

    def __init__(self, name: str):
        self.name = name

    def __str__(self) -> str:
        v_str = complex_to_str(self.voltage)
        s_str = complex_to_str(self.power)
        return f'{self.name}: V = ({v_str}), S = ({s_str})'



import math
from typing import Any, Iterable, List

from .conductor import Conductor


def list_of_lists_to_list(list_of_list: Iterable[List[Any]]) -> List[Any]:
    return [item for sublist in list_of_list for item in sublist]


def geometric_mean_distance(conductors_bundles: List[List[Conductor]]) -> float:
    distance: float = 1
    n: int = 0
    for i, conductors_bundle in enumerate(conductors_bundles):
        compare_conductors_bundles = conductors_bundles[i+1:]
        compare_conductors = list_of_lists_to_list(compare_conductors_bundles)
        for conductor1 in conductors_bundle:
            for conductor2 in compare_conductors:
                d = conductor1.distance(conductor2.position)
                distance *= d if d > 0 else conductor1.gmr
                n += 1
    distance = distance**(1/n)
    return distance


def calc_inductance(conductors_bundles: List[List[Conductor]]) -> float:
    Dm = geometric_mean_distance(conductors_bundles)
    Ds_list = [geometric_mean_distance([c, c]) for c in conductors_bundles]

    L_list = [2e-7*math.log(Dm/Ds) for Ds in Ds_list]
    L = sum(L_list)
    return L



from typing import List

import numpy as np

from .bar import Bar
from .bar.load_bar import LoadBar
from .bar.regulator_bar import RegulatorBar
from .bar.slack_bar import SlackBar
from .line import Line


class Circuit:
    bars: List[Bar]
    lines: List[Line]
    power_base: float
    __y_bus_array: np.ndarray
    __voltage_array: np.ndarray
    __power_esp_array: np.ndarray

    def __init__(self, bars: List[Bar], lines: List[Line], power_base: float = 1):
        self.bars = bars
        self.lines = lines
        self.power_base = power_base
        self.__y_bus_array = self.get_y_bus_array()
        self.__voltage_array = self.get_voltage_array()
        self.__power_esp_array = self.get_power_esp_array()

    def __str__(self) -> str:
        return '\n'.join([str(bar) for bar in self.bars])

    def get_bar_index(self, bar: Bar) -> int:
        for i, b in enumerate(self.bars):
            if b == bar:
                return i
        raise ValueError(f'Bar {bar} not found')

    def get_y_bus_array(self) -> np.ndarray:
        y_bus = np.zeros((len(self.bars), len(self.bars)), dtype=complex)
        for line in self.lines:
            bar1_index = self.get_bar_index(line.bar1)
            bar2_index = self.get_bar_index(line.bar2)
            y_bus[bar1_index, bar2_index] = -line.admittance
            y_bus[bar2_index, bar1_index] = -line.admittance

            y_bus[bar1_index, bar1_index] = sum(
                [l.admittance for l in self.lines if l.bar1 == line.bar1 or l.bar2 == line.bar1])

            y_bus[bar2_index, bar2_index] = sum(
                [l.admittance for l in self.lines if l.bar1 == line.bar2 or l.bar2 == line.bar2])

        return y_bus

    def get_power_esp_array(self) -> np.ndarray:
        power_g = np.zeros((len(self.bars)), dtype=complex)
        power_d = np.zeros((len(self.bars)), dtype=complex)
        for i, bar in enumerate(self.bars):
            if isinstance(bar, LoadBar):
                power_d[i] = bar.power
        power_esp = power_g - power_d
        return power_esp/self.power_base

    def get_voltage_array(self) -> np.ndarray:
        voltages = np.ones(len(self.bars), dtype=complex)
        for i in range(len(voltages)):
            if isinstance(self.bars[i], SlackBar):
                voltages[i] = self.bars[i].voltage
        return voltages

    def update_voltages(self, repeat: int = 1):
        for _ in range(repeat):
            y_bus = self.__y_bus_array
            power_esp = self.__power_esp_array
            voltages = self.__voltage_array

            for i, bar in enumerate(self.bars):
                if isinstance(bar, SlackBar):
                    continue
                I = power_esp[i].conjugate()/voltages[i].conjugate()
                summation = sum([y_bus[i, j]*voltages[j]
                                for j in range(len(voltages)) if i != j])
                voltages[i] = (I - summation)/y_bus[i, i]

            self.__voltage_array = voltages
            for i, bar in enumerate(self.bars):
                bar.voltage = voltages[i]



from schemdraw import Drawing
from schemdraw import elements as elm

from .input import Input


def draw_void():
    drawing = Drawing()

    transistor = drawing.add(elm.transistors.BjtNpn().right())

    drawing.push()
    drawing += elm.Resistor().up().label('Rc')
    drawing += elm.Line().left().length(drawing.unit/2)

    drawing.push()
    drawing += elm.Dot()
    drawing += elm.SourceV().up().reverse().label('Vcc')
    drawing += elm.Ground().left()

    drawing.pop()
    drawing += elm.Line().left().length(drawing.unit/2)
    drawing += elm.Resistor().down().label('Rb')
    drawing += elm.Line().down().length(0.24*drawing.unit)
    drawing += elm.Dot()

    drawing.push()
    drawing += elm.Line().to(transistor.base)

    drawing.pop()
    drawing += elm.Line().left().length(drawing.unit/4)
    drawing += elm.Capacitor().left().reverse().label('Ci')
    drawing += elm.Line().left().length(drawing.unit/4)
    drawing += elm.Dot().label('Vi')

    drawing += elm.Line().at(transistor.emitter).down().length(3*drawing.unit/4)
    drawing += elm.Ground()

    drawing.pop()
    drawing += elm.Dot()
    drawing += elm.Capacitor().right().label('Co')
    drawing += elm.Dot().label('Vo')

    return drawing


def draw(amplifier_input: Input):
    drawing = Drawing()

    transistor = drawing.add(elm.transistors.BjtNpn().right())

    drawing.push()
    drawing += elm.Resistor().up().label(str(amplifier_input.Rc))
    drawing += elm.Line().left().length(drawing.unit/2)

    drawing.push()
    drawing += elm.Dot()
    drawing += elm.SourceV().up().reverse().label(str(amplifier_input.Vcc))
    drawing += elm.Ground().left()

    drawing.pop()
    drawing += elm.Line().left().length(drawing.unit/2)
    drawing += elm.Resistor().down().label(str(amplifier_input.Rb))
    drawing += elm.Line().down().length(0.24*drawing.unit)
    drawing += elm.Dot()

    drawing.push()
    drawing += elm.Line().to(transistor.base)

    drawing.pop()
    drawing += elm.Line().left().length(drawing.unit/4)
    drawing += elm.Capacitor().left().reverse().label('Ci')
    drawing += elm.Line().left().length(drawing.unit/4)
    drawing += elm.Dot().label('Vi')

    drawing += elm.Line().at(transistor.emitter).down().length(3*drawing.unit/4)
    drawing += elm.Ground()

    drawing.pop()
    drawing += elm.Dot()
    drawing += elm.Capacitor().right().label('Co')
    drawing += elm.Dot().label('Vo')

    return drawing



from __future__ import annotations

import cmath
import math
from typing import Union

from pyelectric.ac.phasor import Phasor


class Power:
    complex_power: complex
    factor: float
    is_late: bool
    phase: float

    def __init__(
        self,
        complex_power: Union[complex, Phasor] = None, *,
        active: float = None,
        reactive: float = None,
        factor: float = None,
        apparent: float = None,
        is_late: bool = None,
    ) -> None:
        if factor == 1:
            is_late = False
        if complex_power is not None:
            if isinstance(complex_power, Phasor):
                complex_power = complex_power.value
            self.complex_power = complex_power
        elif (active, reactive).count(None) == 0:
            self.complex_power = complex(active, reactive)
        elif (factor, apparent, is_late).count(None) == 0:
            theta = math.acos(factor)
            P, Q = apparent*math.cos(theta), apparent*math.sin(theta)
            self.complex_power = complex(P, Q if is_late else -Q)
        elif (active, apparent, is_late).count(None) == 0:
            P, Q = active, apparent*math.sin(math.acos(active/apparent))
            self.complex_power = complex(P, Q if is_late else -Q)
        elif (reactive, apparent, is_late).count(None) == 0:
            theta = math.asin(reactive/apparent)
            P, Q = apparent*math.cos(theta), reactive
            self.complex_power = complex(P, Q if is_late else -Q)
        elif (active, factor, is_late).count(None) == 0:
            theta = math.acos(factor)
            apparent = active/factor
            P, Q = active, apparent*math.sin(theta)
            self.complex_power = complex(P, Q if is_late else -Q)
        elif (reactive, factor, is_late).count(None) == 0:
            theta = math.acos(factor)
            apparent = reactive/math.sin(theta)
            P, Q = apparent*math.cos(theta), reactive
            self.complex_power = complex(P, Q if is_late else -Q)
        else:
            raise ValueError('insufficient power information')

    @property
    def active(self) -> float:
        return self.complex_power.real

    @property
    def reactive(self) -> float:
        return self.complex_power.imag

    @property
    def apparent(self) -> float:
        return math.sqrt(self.active**2 + self.reactive**2)

    @property
    def factor(self) -> float:
        return self.active/self.apparent

    @property
    def is_late(self) -> bool:
        return self.reactive > 0

    @property
    def phase(self) -> float:
        return cmath.phase(self.complex_power)

    def print_powers(self) -> None:
        print(f'Active: {self.active:.2f}')
        print(f'Reactive: {self.reactive:.2f}')
        print(f'Apparent: {self.apparent:.2f}')
        print(f'Factor: {self.factor:.2f}')
        print(f'Phase: {math.degrees(self.phase):.2f}°')
        print(f'Is late: {self.is_late}')

    def __str__(self) -> str:
        return f'{self.active:.2f} + {self.reactive:.2f}j'

    def __repr__(self) -> str:
        return self.__str__()

    def __mod__(self, other: Power) -> Power:
        return Power(self.complex_power % other.complex_power)

    def __add__(self, other: Union[Power, Phasor, float]) -> Power:
        if isinstance(other, Power):
            return Power(self.complex_power + other.complex_power)
        elif isinstance(other, Phasor):
            return Power(self.complex_power + other.value)
        else:
            return Power(self.complex_power + other)

    def __sub__(self, other: Union[Power, Phasor, float]) -> Power:
        if isinstance(other, Power):
            return Power(self.complex_power - other.complex_power)
        elif isinstance(other, Phasor):
            return Power(self.complex_power - other.value)
        else:
            return Power(self.complex_power - other)

    def __mul__(self, other: Union[Power, Phasor, float]) -> Power:
        if isinstance(other, Power):
            return Power(self.complex_power * other.complex_power)
        elif isinstance(other, Phasor):
            return Power(self.complex_power * other.value)
        else:
            return Power(self.complex_power * other)

    def __truediv__(self, other: Union[Power, Phasor, float]) -> Power:
        if isinstance(other, Power):
            return Power(self.complex_power / other.complex_power)
        elif isinstance(other, Phasor):
            return Power(self.complex_power / other.value)
        else:
            return Power(self.complex_power / other)



from schemdraw import Drawing
from schemdraw import elements as elm

from .input import Input


def draw_void():
    drawing = Drawing()

    transistor = drawing.add(elm.transistors.JFetP().right().reverse())

    drawing += elm.SourceV().at(transistor.drain).up().label('Vdd').reverse()
    drawing += elm.Ground().left()

    drawing += elm.Line().left().at(transistor.gate).length(drawing.unit/4)

    drawing.push()
    drawing += elm.Dot()
    drawing += elm.Resistor().down().label('Rg')
    drawing += elm.Ground()

    drawing.pop()
    drawing += elm.Capacitor().left()
    drawing += elm.Dot().label('Vi')

    drawing += elm.Dot().at(transistor.source).down()

    drawing.push()
    drawing += elm.Resistor().down().label('Rs').length(drawing.unit*0.857)
    drawing += elm.Ground()

    drawing.pop()
    drawing += elm.Capacitor().right().label('C2')
    drawing += elm.Dot().label('Vo')

    return drawing


def draw(amplifier_input: Input):
    Vdd = str(amplifier_input.Vdd)
    Rg = str(amplifier_input.Rg)
    Rs = str(amplifier_input.Rs)

    drawing = Drawing()

    transistor = drawing.add(elm.transistors.JFetP().right().reverse())

    drawing += elm.SourceV().at(transistor.drain).up().label(Vdd).reverse()
    drawing += elm.Ground().left()

    drawing += elm.Line().left().at(transistor.gate).length(drawing.unit/4)

    drawing.push()
    drawing += elm.Dot()
    drawing += elm.Resistor().down().label(Rg)
    drawing += elm.Ground()

    drawing.pop()
    drawing += elm.Capacitor().left()
    drawing += elm.Dot().label('Vi')

    drawing += elm.Dot().at(transistor.source).down()

    drawing.push()
    drawing += elm.Resistor().down().label(Rs).length(drawing.unit*0.857)
    drawing += elm.Ground()

    drawing.pop()
    drawing += elm.Capacitor().right().label('C2')
    drawing += elm.Dot().label('Vo')

    return drawing



from . import Bar


class RegulatorBar(Bar):
    voltage_module: float
    active_power: float

    def __init__(self, name: str, voltage_module: float, active_power: float):
        super().__init__(name)
        self.voltage_module = voltage_module
        self.active_power = active_power



from __future__ import annotations

import cmath
import math
from typing import Union


class Phasor:
    value: complex
    decimal_places: int = 2
    angle_in_degree: bool = True

    def __init__(self, r: Union[float, complex], theta: float = None) -> None:
        if theta is None:
            r, theta = cmath.polar(r)
        self.value = cmath.rect(r, theta)

    @property
    def abs(self):
        return abs(self.value)

    @property
    def phase(self):
        return cmath.phase(self.value)

    @abs.setter
    def abs(self, value: float):
        self.value = cmath.rect(value, self.phase)

    @phase.setter
    def phase(self, value: float):
        self.value = cmath.rect(self.abs, value)

    def conjugate(self) -> Phasor:
        return Phasor(self.value.conjugate())

    def __str__(self) -> str:
        r, theta = cmath.polar(self.value)
        r = round(r, self.decimal_places)
        if self.angle_in_degree:
            theta = math.degrees(theta)
            theta = round(theta, self.decimal_places)
            return f"{r} ∠ {theta}°"
        theta = round(theta, self.decimal_places)
        return f"{r} ∠ {theta}"

    def __repr__(self) -> str:
        return self.__str__()

    def __add__(self, other: Union[Phasor, complex]) -> Phasor:
        return Phasor(self.value + (other.value if isinstance(other, Phasor) else other))

    def __sub__(self, other: Union[Phasor, complex]) -> Phasor:
        return Phasor(self.value - (other.value if isinstance(other, Phasor) else other))

    def __mul__(self, other: Union[Phasor, complex]) -> Phasor:
        return Phasor(self.value * (other.value if isinstance(other, Phasor) else other))

    def __truediv__(self, other: Union[Phasor, complex]) -> Phasor:
        return Phasor(self.value / (other.value if isinstance(other, Phasor) else other))

    def __pow__(self, other: Union[Phasor, complex]) -> Phasor:
        return Phasor(self.value ** (other.value if isinstance(other, Phasor) else other))

    def __floordiv__(self, other: Union[Phasor, complex]) -> Phasor:
        return Phasor(1 / ((1 / self.value) + (1 / other.value if isinstance(other, Phasor) else other)))

    def __eq__(self, other: Union[Phasor, complex]) -> bool:
        return self.value == (other.value if isinstance(other, Phasor) else other)
